HALLUCINATION_PROMPT_TEMPLATE = """
In this task, you will be presented with a query, a reference text and an answer. The answer is
generated to the question based on the reference text. The answer may contain false information. You
must use the reference text to determine if the answer to the question contains false information,
if the answer is a hallucination of facts. Your objective is to determine whether the answer text
contains factual information and is not a hallucination. A 'hallucination' refers to
an answer that is not based on the reference text or assumes information that is not available in
the reference text. "hallucinated" indicates that the answer
provides factually inaccurate information to the query based on the reference text. "factual"
indicates that the answer to the question is correct relative to the reference text, and does not
contain made up information. Please read the query and reference text carefully before determining
your response.

[BEGIN DATA]
************
[Query]: {input}
************
[Reference text]: {reference}
************
[Answer]: {output}
************
[END DATA]

Is the answer above factual or hallucinated based on the query and reference text? 
Your response should be a single word: either "factual" or "hallucinated", and it should not include any other text or characters.
"""

QA_REFERENCE_PROMPT_TEMPLATE = """
You are given a question, an answer and reference text. You must determine whether the
given answer correctly answers the question based on the reference text. Here is the data:
[BEGIN DATA]
************
[Question]: {input}
************
[Reference]: {reference}
************
[Answer]: {output}
[END DATA]

Your response must be a single word, either "correct" or "incorrect", and should not contain any text or characters aside from that word.
"correct" means that the question is correctly and fully answered by the answer.
"incorrect" means that the question is not correctly or only partially answered by the answer.
"""

MODEL_GRADING_QA_PROMPT_TEMPLATE = """
You are grading output according to a user-specified rubric. If the statement in the rubric is true, then the output passes the test.

[EXAMPLES]
************
[Output]: Hello world
************
[Rubric]: Content contains a greeting
************
correct

************
[Output]: Avast ye swabs, repel the invaders!
************
[Rubric]: Does not speak like a pirate
************
incorrect
[END EXAMPLES]

[BEGIN DATA]
[Output]: {output}
************
[Rubric]: {assertion}
************
[END DATA]

Your response must be a single word, either "correct" or "incorrect", and should not contain any text or characters aside from that word.
"correct" means that the output meets the criteria specified in the rubric.
"incorrect" means that the output does not meet the criteria specified in the rubric.
"""

SUMMARIZATION_PROMPT_TEMPLATE = """
You are comparing the summary text and it's original document and trying to determine if the summary is good. Here is the data:
[BEGIN DATA]
************
[Summary]: {output}
************
[Original Document]: {input}
[END DATA]

Compare the Summary above to the Original Document and determine if the Summary is comprehensive, concise, coherent, 
and independent relative to the Original Document. 
Your response must be a single word, either "good" or "bad", and should not contain any text or characters aside from that. 
"bad" means that the Summary is not comprehensive, concise, coherent, and independent relative to the Original Document. 
"good" means the Summary is comprehensive, concise, coherent, and independent relative to the Original Document.
"""

HUMAN_VS_AI_PROMPT_TEMPLATE = """
You are comparing a human ground truth answer from an expert to an answer from an AI model.
Your goal is to determine if the AI answer correctly matches, in substance, the human answer.
[BEGIN DATA]
************
[Question]: {input}
************
[Human Ground Truth Answer]: {reference}
************
[AI Answer]: {output}
************
[END DATA]

Compare the AI answer to the human ground truth answer, if the AI correctly answers the question, then the AI answer is "correct". 
If the AI answer is longer but contains the main idea of the Human answer please answer "correct". 
If the AI answer divergences or does not contain the main idea of the human answer, please answer "incorrect".
Your response must be a single word, either "correct" or "incorrect", and should not contain any text or characters aside from that. 
"""



from enum import StrEnum

from pydantic import BaseModel, ConfigDict


class EndpointsEnum(StrEnum):
    OPENAI = "openai"
    OPENAI_COMPATIBLE = "openai_compatible"
    TG_CHATBOT = "tg_chatbot"
    ECHO = "echo"
    CC_PROMPT_LLM = "cc_prompt_llm"
    CC_SS = "cc_ss"


# NOTE: Not every Endpoint needs to use custom config through extension of the EndpointConfig
class EndpointConfig(BaseModel):
    model_config = ConfigDict(extra="allow")
    kind: EndpointsEnum = EndpointsEnum.OPENAI
    model: str = "gpt-4o-mini"
    temperature: float | None = None
    max_tokens: int | None = None



import sys
from argparse import ArgumentParser

from contextcheck.endpoints.endpoint_config import EndpointConfig
from contextcheck.generators.generate_questions import QuestionsGenerator
from contextcheck.generators.utils import import_class_from_string


def generate_questions(
    output_file: str,
    wrapper_class_path: str,
    num_topics: int,
    questions_per_topic: int,
    llm_model_provider: str,
    llm_model: str,
) -> None:
    api_wrapper = import_class_from_string(wrapper_class_path)

    endpoint_config = EndpointConfig(
        kind="openai_compatible", provider=llm_model_provider, model=llm_model
    )

    generator_args = {
        "api_wrapper": api_wrapper(),
        "questions_generator_endpoint_config": endpoint_config,
    }

    if num_topics:
        generator_args["num_topics"] = num_topics
    if questions_per_topic:
        generator_args["questions_per_topic"] = questions_per_topic

    generator = QuestionsGenerator.model_validate(generator_args)
    generator.save_to_yaml(output_file)


def main():
    parser = ArgumentParser(
        prog="questions_generator.py", description="Refer to readme for more information."
    )
    parser.add_argument(
        "--output-file",
        type=str,
        required=True,
        help="Output path and filename to save the generated questions. "
        "Example: 'my_folder/my_file.yaml'. If folder does not exist, it will be created.",
    )
    parser.add_argument(
        "--wrapper-class-path",
        type=str,
        required=True,
        help="Full path to the API wrapper class definition. "
        "Example: For a class 'ClassName' defined in file 'module/submodule.py' use format 'module.submodule.ClassName'.",
    )
    parser.add_argument(
        "--num-topics",
        type=int,
        default=10,
        help="Number of topics generated for each file. Default is 10",
    )
    parser.add_argument(
        "--questions-per-topic",
        type=int,
        default=3,
        help="Number of questions generated per each topic. Default is 3.",
    )
    parser.add_argument(
        "--llm-model-provider",
        type=str,
        default="ChatOpenAI",
        help="LLM model provider. Default one is OpenAI. Please refer to "
        "https://python.langchain.com/v0.2/docs/integrations/chat/ for more information",
    )
    parser.add_argument(
        "--llm-model",
        type=str,
        default="gpt-4o-mini",
        help="LLM model to use. Default one is GPT-4o-Mini."
        "Please also ensure that the chosen model is compatible with a given provider.",
    )

    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])

    generate_questions(
        output_file=args.output_file,
        wrapper_class_path=args.wrapper_class_path,
        num_topics=args.num_topics,
        questions_per_topic=args.questions_per_topic,
        llm_model_provider=args.llm_model_provider,
        llm_model=args.llm_model,
    )


if __name__ == "__main__":
    main()



from contextcheck.connectors.connector import ConnectorBase


class ConnectorEcho(ConnectorBase):
    """Sends the same data as received."""

    def send(self, data: dict) -> dict:
        return data



import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import Literal

from loguru import logger
from pydantic import BaseModel, field_validator, model_validator

from contextcheck.executors.executor import Executor
from contextcheck.interfaces.interface import InterfaceBase
from contextcheck.interfaces.interface_output_file import InterfaceOutputFile
from contextcheck.interfaces.interface_tui import InterfaceTUI
from contextcheck.models.models import TestScenario


class TestsRouter(BaseModel):
    output_type: Literal["console", "file"]
    filename: list[Path] | None = []
    folder: Path | None = None
    output_folder: Path | None = None
    exit_on_failure: bool = False
    global_test_timestamp: str | None = None
    aggregate_results: bool = False
    show_time_statistics: bool = False

    @model_validator(mode="before")
    @classmethod
    def initialize_paths(cls, data: dict) -> dict:
        if "output_folder" in data and data["output_folder"]:
            data["output_folder"] = Path(data["output_folder"])
        if "folder" in data and data["folder"]:
            data["folder"] = Path(data["folder"])
        if "filename" in data and data["filename"]:
            data["filename"] = [Path(file) for file in data["filename"]]
        return data

    @field_validator("filename")
    @classmethod
    def check_filename(cls, value: list[Path] | None) -> list[Path]:
        if value:
            invalid_files = [file for file in value if not file.is_file()]
            if invalid_files:
                raise ValueError(
                    f'Files {", ".join(map(str, invalid_files))} do not exist. Check --filename'
                    " argument."
                )
        return value or []

    @field_validator("folder")
    @classmethod
    def check_folder(cls, value: Path | None) -> Path | None:
        if value and not value.is_dir():
            raise ValueError(f'Folder "{value}" does not exist. Check --folder argument.')
        return value

    @field_validator("output_folder")
    @classmethod
    def ensure_output_folder_exists(cls, value: Path | None) -> Path | None:
        if value and not value.exists():
            value.mkdir(parents=True)
        return value

    def model_post_init(self, __context) -> None:
        if not self.global_test_timestamp:
            now = datetime.now(timezone.utc)
            self.global_test_timestamp = str(datetime.timestamp(now))

    def run_tests(self):
        scenario_results = []
        executors: list[Executor] = []
        type_map = {"console": InterfaceTUI, "file": InterfaceOutputFile}
        interface_type = type_map[self.output_type]

        filenames = self.filename or []
        if self.folder:
            filenames.extend(self.folder.glob("*.yaml"))

        if not filenames:
            logger.warning("No test scenario to run.")
            return

        # TODO: Potential place to increase performance by running several scenarios at once
        for filename in filenames:
            executor, scenario_result, early_stop = self._run_test_scenario(
                filename, interface_type
            )
            executors.append(executor)
            scenario_results.append(scenario_result)
            if self.exit_on_failure and early_stop:
                break

        for executor in executors:
            executor.summary(
                output_folder=str(self.output_folder) if self.output_folder else None,
                global_test_timestamp=self.global_test_timestamp,
                aggregate_results=self.aggregate_results,
                show_time_statistics=self.show_time_statistics,
            )

        if self.exit_on_failure and not all(scenario_results):
            sys.exit(1)

    def _run_test_scenario(
        self, filename: Path, interface_type: InterfaceBase
    ) -> tuple[Executor, bool | None, bool]:
        ui = interface_type(test_scenario_filename=str(filename))  # type: ignore
        ts = TestScenario.from_yaml(ui.get_scenario_path())

        executor = Executor(ts, ui=ui, exit_on_failure=self.exit_on_failure)
        scenario_result = executor.run_all()

        return executor, scenario_result, executor.early_stop



from pydantic import Field, model_serializer, model_validator

from contextcheck.connectors.connector_openai_compatible import ConnectorOpenAICompatible
from contextcheck.endpoints.endpoint import EndpointBase, EndpointConfig
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase, ResponseStats


class EndpointOpenAICompatibleConfig(EndpointConfig):
    provider: str = "ChatOpenAI"


class EndpointOpenAICompatible(EndpointBase):
    config: EndpointOpenAICompatibleConfig = Field(
        default_factory=EndpointOpenAICompatibleConfig,
        description="Configuration for endpoints compatible with OpenAI through langchain"
        ". Please see `ConnectorOpenAICompatible` for more info",
    )
    connector: ConnectorOpenAICompatible = ConnectorOpenAICompatible()

    def model_post_init(self, __context) -> None:
        # TODO: Check why self.connector.model_copy won't work - hint: field validator of connector
        self.connector = ConnectorOpenAICompatible(
            model=self.config.model,
            provider=self.config.provider,
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
            **self.connector.model_dump(exclude={"model", "provider", "temperature", "max_tokens"})
        )

    class RequestModel(RequestBase):
        @model_serializer
        def serialize(self) -> dict:
            return {"role": "user", "content": self.message}

    class ResponseModel(ResponseBase):
        @model_validator(mode="before")
        @classmethod
        def from_dict(cls, data: dict) -> dict:
            data["message"] = data["choices"][0]["message"]["content"]

            # Attempt to collect stats
            if (
                "usage" in data
                and "prompt_tokens" in data["usage"]
                and "completion_tokens" in data["usage"]
                and "total_tokens" in data["usage"]
            ):
                data["stats"] = ResponseStats(
                    tokens_request=data["usage"]["prompt_tokens"],
                    tokens_response=data["usage"]["completion_tokens"],
                    tokens_total=data["usage"]["total_tokens"],
                )
            return data



import json
import os
from datetime import datetime, timezone
from pathlib import Path

import fsspec
from loguru import logger

from contextcheck.executors.executor import Executor
from contextcheck.interfaces.interface import InterfaceBase


def create_output_path(
    file_path: str, output_folder: str, suffix: str = ""
) -> tuple[str, datetime]:
    filename = os.path.basename(file_path)  # type: ignore
    filename = filename.split(".")[0] + suffix
    date_now = datetime.now(timezone.utc)
    output_path = f"{output_folder}/{filename}_{date_now.strftime('%Y-%m-%d_%H-%M-%S')}.json"

    return output_path, date_now


class InterfaceOutputFile(InterfaceBase):
    test_scenario_filename: str | None = None

    def get_scenario_path(self) -> Path:
        return Path(self.test_scenario_filename or "")

    def summary(
        self, executor: Executor, output_folder: str, global_test_timestamp: str, **kwargs
    ) -> None:
        output_path, date_now = create_output_path(
            file_path=self.test_scenario_filename, output_folder=output_folder
        )

        res = executor.test_scenario.model_dump()
        res["global_test_timestamp"] = global_test_timestamp
        res["test_timestamp"] = str(datetime.timestamp(date_now))

        if kwargs.get("aggregate_results", False):
            report = self._create_a_summary_report(executor=executor)
            res["report"] = report

        if kwargs.get("show_time_statistics", False):
            time_statistics = self._create_time_statistics(executor=executor)
            res["time_statistics"] = time_statistics

        res = json.dumps(res, indent=4)
        FileHandler(output_path).write_file(res)

    def report_results(
        self, executor: Executor, output_folder: str, global_test_timestamp: str
    ) -> None:
        output_path, date_now = create_output_path(
            file_path=self.test_scenario_filename, output_folder=output_folder, suffix="_report"
        )

        res = {}
        res["global_test_timestamp"] = global_test_timestamp
        res["test_timestamp"] = str(datetime.timestamp(date_now))

        scenario_results = self._create_a_summary_report(executor=executor)
        res["report"] = scenario_results

        res = json.dumps(res, indent=4)
        FileHandler(output_path).write_file(res)

    def report_time(
        self, executor: Executor, output_folder: str, global_test_timestamp: str
    ) -> None:
        output_path, date_now = create_output_path(
            file_path=self.test_scenario_filename,
            output_folder=output_folder,
            suffix="_time_statistics",
        )

        res = {}
        res["global_test_timestamp"] = global_test_timestamp
        res["test_timestamp"] = str(datetime.timestamp(date_now))

        scenario_results = self._create_a_summary_report(executor=executor)
        res["time_statistics"] = scenario_results

        res = json.dumps(res, indent=4)
        FileHandler(output_path).write_file(res)


class FileHandler:
    def __init__(self, file_path):
        self.file_path = file_path

    def write_file(self, content: str):
        with fsspec.open(self.file_path, "w") as f:
            f.write(content)  # type: ignore
        logger.info(f"File written to {self.file_path}")

    def read_file(self):
        with fsspec.open(self.file_path, "r") as f:
            content = f.read()  # type: ignore
        logger.info(f"File read from {self.file_path}")
        return content



import requests

from contextcheck.connectors.connector import ConnectorBase


class ConnectorHTTP(ConnectorBase):
    additional_headers: dict = {}
    url: str
    method: str = "POST"
    timeout: int = 30

    def send(
        self, data: dict, json: bool = True, params: dict | None = None, timeout: int | None = None
    ) -> dict:
        response = requests.request(
            method=self.method,
            url=self.url,
            json=data if json else None,
            data=data if not json else None,
            params=params,
            timeout=timeout or self.timeout,
            headers=self.additional_headers,
        )
        response.raise_for_status()
        return response.json()



import json
import os

import nltk
import yaml
from gensim import corpora
from gensim.models import LdaModel
from gensim.utils import simple_preprocess
from nltk.corpus import stopwords
from pydantic import BaseModel

from contextcheck.endpoints.endpoint import EndpointBase
from contextcheck.endpoints.endpoint_config import EndpointConfig
from contextcheck.endpoints.factory import factory as endpoint_factory
from contextcheck.generators.endpoint_wrapper import RagApiWrapperBase
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase

PROMPT_SINGLE_QUESTION = """Generate question basing on the content chunks above. Only respond with the question, do not provide answers. 
Put question in a separate JSONL entity. Use the following format:
{"question": "<QUESTION>"}
"""

PROMPT_QUESTIONS = """Generate {num_questions} questions basing on the content chunks above. Only respond with the questions, do not provide answers. 
Put each question in a separate JSONL entity. Use the following format:
{{"question": "<QUESTION>"}}
"""


class QuestionsGenerator(BaseModel):
    num_topics: int = 10
    questions_per_topic: int = 3
    words_per_topic: int = 30

    api_wrapper: RagApiWrapperBase
    questions_generator_endpoint_config: EndpointConfig
    llm_endpoint: EndpointBase | None = None

    stop_words: set = set()
    generated_questions: list[str] = []

    def model_post_init(self, __context) -> None:
        nltk.download("stopwords")
        self.stop_words = set(stopwords.words(["english", "spanish"]))
        self.llm_endpoint = endpoint_factory(self.questions_generator_endpoint_config)

    def preprocess_text(self, text: str):
        result = []
        for token in simple_preprocess(text, deacc=True):
            if token not in self.stop_words and len(token) > 3:
                result.append(token)
        return result

    def get_topic_lists_from_chunks(self, documents: list[str]) -> list[list[str]]:
        processed_documents = [self.preprocess_text(doc) for doc in documents]

        dictionary = corpora.Dictionary(processed_documents)
        corpus = [dictionary.doc2bow(doc) for doc in processed_documents]

        lda_model = LdaModel(corpus, num_topics=self.num_topics, id2word=dictionary, passes=15)

        topics = lda_model.print_topics(num_words=self.words_per_topic)

        topics_ls = []
        for topic in topics:
            words = topic[1].split("+")
            topic_words = [word.split("*")[1].replace('"', "").strip() for word in words]
            topics_ls.append(topic_words)

        return topics_ls

    def _prepare_request(self, chunks: list[dict]) -> RequestBase:
        if self.questions_per_topic > 1:
            prompt = PROMPT_QUESTIONS.format(num_questions=self.questions_per_topic)
        else:
            prompt = PROMPT_SINGLE_QUESTION

        chunks_joined = "\n---\n".join([chunk["chunk"] for chunk in chunks])
        message = f"{chunks_joined}.\n\n------\n{prompt}"
        return RequestBase(message=message)

    def _parse_response(self, response: ResponseBase, split_characters: str = "\n") -> list[str]:
        """Split the response by newlines and extract the question from each JSON entity.
        Return a list of questions."""
        msg = response.message
        questions = msg.split(split_characters)
        parsed_questions = []
        for q in questions:
            try:
                parsed_questions.append(json.loads(q)["question"])
            except:
                continue
        return parsed_questions

    def generate(self) -> list[str]:
        documents = self.api_wrapper.list_documents()

        questions = []

        for document in documents:
            chunks = self.api_wrapper.get_document_chunks(document["id"])
            document_topics = self.get_topic_lists_from_chunks(chunks)

            topic_questions = []
            for topic_num, topic in enumerate(document_topics):
                print(f"Generating questions for document {document['name']} topic {topic_num}.")
                # Get chunks from semantic db with space separated topic words
                topic_chunks = self.api_wrapper.query_semantic_db(" ".join(topic))

                prompt = self._prepare_request(topic_chunks)
                response = self.llm_endpoint.send_request(prompt)
                topic_questions += self._parse_response(response)

            questions.append({"document": document["name"], "questions": topic_questions})

        self.generated_questions = questions
        return questions

    def save_to_yaml(self, filepath: str):
        if not self.generated_questions:
            self.generate()

        dirs = os.path.dirname(filepath)
        if dirs:
            os.makedirs(dirs, exist_ok=True)

        with open(filepath, "w") as f:
            yaml.dump({"questions": self.generated_questions}, stream=f, width=200)



from pydantic import BaseModel, ConfigDict, Field

from contextcheck.connectors.connector import ConnectorBase
from contextcheck.endpoints.endpoint_config import EndpointConfig
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase


class EndpointBase(BaseModel):
    model_config = ConfigDict(extra="allow")
    connector: ConnectorBase = Field(
        default_factory=ConnectorBase,
        description="A connector responsible for connection to llm/rag",
    )
    config: EndpointConfig = Field(
        default_factory=EndpointConfig, description="Configuration for llm/rag"
    )

    class RequestModel(RequestBase):
        pass

    class ResponseModel(ResponseBase):
        pass

    def send_request(self, req: RequestBase) -> ResponseBase:
        req = self.RequestModel(**req.model_dump())
        with self.connector as c:
            response_dict = c.send(req.model_dump())
        response_dict.update({"config": self.config})
        response = self.ResponseModel.model_validate(response_dict)

        # Add connector stats to response stats:
        response.stats = response.stats.model_copy(update=c.stats.model_dump())

        return response



import os

import dotenv
import requests

from contextcheck.generators.endpoint_wrapper import RagApiWrapperBase

dotenv.load_dotenv()


class ContextClueApiWrapper(RagApiWrapperBase):
    endpoint_base_url: str = os.environ["ENDPOINT_BASE_URL"]
    header_key: str = os.environ["HEADER_KEY"]
    header_value: str = os.environ["HEADER_VALUE"]

    @property
    def headers(self) -> dict[str, str]:
        return {self.header_key: self.header_value}

    def list_documents(self) -> list[dict[str, str]]:
        response = requests.get(
            f"{self.endpoint_base_url}/documents/", headers=self.headers, timeout=self.timeout
        )
        response.raise_for_status()

        return [
            {"id": document["id"], "name": document["name"]}
            for document in response.json()["documents"]
        ]

    def get_document_chunks(self, document_id: str) -> list[str]:
        response = requests.get(
            f"{self.endpoint_base_url}/documents/{document_id}/",
            headers=self.headers,
            timeout=self.timeout,
        )
        response.raise_for_status()

        return response.json()["document"]["chunks"]

    def query_semantic_db(self, query: str, **kwargs) -> list[dict]:
        response = requests.post(
            f"{self.endpoint_base_url}/semantic_search/get_relevant_documents",
            json={
                "query": query,
                "top_k": kwargs.get("top_k", 3),
                "alpha": kwargs.get("alpha", 0.75),
                "use_ranker": kwargs.get("use_ranker", True),
            },
            headers=self.headers,
            timeout=self.timeout,
        )
        response.raise_for_status()
        chunks = response.json()["relevant_documents"]["collection_retriever_entries"]
        return chunks

    def query_qa(self, query: str, **kwargs) -> dict:
        response = requests.post(
            f"{self.endpoint_base_url}/qa/ask",
            json={
                "query": query,
                "alpha": kwargs.get("alpha", 0.75),
                "rag_config": {"temperature": 0, "llm": "openai", "top_k": 3},
            },
            headers=self.headers,
            timeout=self.timeout,
        )
        response.raise_for_status()
        answer = response.json()
        return answer



from pydantic import BaseModel, ConfigDict, model_validator


class ResponseStats(BaseModel):
    model_config = ConfigDict(extra="allow")
    tokens_request: int | None = None
    tokens_response: int | None = None
    tokens_total: int | None = None


error_message = """You see this error because of incorrect response parsing.
Your response's dictionary format causes validation error.
In order to avoid this error, you can `parse_response_as_json: true` either 
in your config.default_request or in the request step. Example:

config:
    endpoint_under_test:
        kind: echo
    default_request:
        parse_response_as_json: true  # Set globally for all tests in this file

steps:
  - name: test parsing as json
    request: '{"Hello": "World"}'
    parse_response_as_json: true  # Set for this test only
"""


class ResponseBase(BaseModel):
    model_config = ConfigDict(extra="allow")
    message: str | None = None
    stats: ResponseStats = ResponseStats()

    @model_validator(mode="before")
    @classmethod
    def from_obj(cls, obj: dict) -> dict:
        if (
            obj.get("parse_response_as_json", False)
            and obj.get("message")
            and isinstance(obj["message"], dict)
        ):
            raise ValueError(error_message)
        return obj



import os
from pathlib import Path

import yaml
from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from contextcheck.generators.endpoint_wrapper import RagApiWrapperBase
from contextcheck.loaders.yaml import load_yaml_file


class DocumentQuestions(BaseModel):
    model_config = ConfigDict(extra="ignore")

    document: str = Field(description="A document from which a question is derived?")
    questions: list[str] = Field(description="A list of questions regarding a document")


class AnswerGenerator(BaseModel):
    top_k: int = 3
    collection_name: str = "default"
    questions_file: Path
    api_wrapper: RagApiWrapperBase
    questions: list[DocumentQuestions] = Field(
        default_factory=list, description="A list of nested questions regarding a document"
    )
    alpha: float = 0.75
    use_ranker: bool = True
    debug: bool = False

    def model_post_init(self, __context):
        questions_dict = load_yaml_file(self.questions_file)["questions"]
        self.questions = TypeAdapter(list[DocumentQuestions]).validate_python(questions_dict)

    def generate(self) -> dict:
        """
        Generate answers for the given questions.

        Returns:
            dict: A dictionary containing the questions and their answers.
        """

        qa_data = {"QA": []}

        for document_questions in self.questions:
            current_document = document_questions.document
            entry = {"document": current_document, "qa": []}
            list_of_questions = document_questions.questions
            print("Generating answers for document:", current_document)
            for idx, question in enumerate(list_of_questions):
                answer = self.api_wrapper.query_qa(
                    question, use_ranker=self.use_ranker, top_k=self.top_k, alpha=self.alpha
                )
                qa_item = {
                    "question": question,
                    "answer": answer["result"],
                }

                if self.debug:
                    qa_item["chunks_and_documents"] = [
                        [
                            {
                                "chunk": answer["chunk"],
                                "document": answer["metadata"]["document_name"],
                            }
                        ]
                        for answer in answer.get("relevant_documents", {}).get(
                            "collection_retriever_entries", []
                        )
                    ]

                print(f"Processing {idx + 1}/{len(list_of_questions)} questions")
                entry["qa"].append(qa_item)
            qa_data["QA"].append(entry)

        return qa_data

    def save_to_yaml(self, filepath: str):
        """
        Save the generated questions to a YAML file.

        Args:
            filepath (str): The path to the output file.

        Returns:
            None
        """
        qa = self.generate()
        dirs = os.path.dirname(filepath)
        if dirs:
            os.makedirs(dirs, exist_ok=True)

        with open(filepath, "w") as f:
            yaml.dump(qa, stream=f, width=200)
